#!/usr/bin/env python
"""
author: Jun Ding
since: 01/07/2014
usage: python extend.py <.motifs> <.mc> <hits> <pos> <seq> <out>
"""
import pdb,sys,os,StatTest,gc
from StatTest import *
gc.enable()


########################################################################
#modules 

#=======================================================================
# This is used to get the targets of input motif m
def motTarget(m,tfbs):
	mt=[]
	# based on tfbs (motif hit file), we will get the targts of given motifs
	for i in range(len(tfbs)):
		if m in tfbs[i]:
			mt.append(i)
	return mt
	
#=======================================================================
# This is used to read in the .motifs file
def readMot(fname):
	f=open(fname,'r')
	lf=f.readlines()
	f.close()
	lf="".join(lf)
	lf=lf.split(">")[1:]
	for i in range(len(lf)):
		lf[i]=lf[i].split("\n")
		lf[i]=[item for item in lf[i] if item!=""]
		lf[i]=[lf[i][0].split()[0],len(lf[i])-1,lf[i][0].split()[-1]]
	mot=lf
	return mot

#=======================================================================
# read in motif hits and motif hits position 
# genearally, the hits information of motifs 
def readMotHit(fname):
	f=open(fname,'r')
	lf=f.readlines()
	f.close()
	lf=[item.strip().split() for item in lf]
	return lf 

#=======================================================================
# read in the original sequence

def readSeq(fname):
	f=open(fname,'r')
	lf=f.readlines()
	f.close()
	lf="".join(lf)
	lf=lf.split(">")[1:]
	lf=[item.split("\n") for item in lf]
	lf=[item for item in lf if item!=""]
	lf=["".join(item[1:]) for item in lf]
	return lf	

#=======================================================================
# targets of motifs in motif modules (all modules having the motifs)
# m: given motif
# dm: dictionary of m->targets
# dmm: dictionary of m->targets (of modules having the motifs)

def motTargetMC(m,dm,dmm):
	A=dm[m]
	ml=dmm[m]
	d={}
	for i in ml:
		if i!=m:
			ti=dm[i]
			for j in ti:
				if d.has_key(j)==False:
					d[j]=0
					
	T=[]
	for i in A:
		if d.has_key(i)==True:
			T.append(i)
	return T


#=======================================================================
# get_matrix based on input Xmer hits
# A: motif Hits

def get_matrix(A):
	dN={"A":0,"C":1,"G":2,"T":3}
	pc=0.01
	mat=[]
	x=len(A[0])
	for i in range(x):
		N=[0,0,0,0]
		for j in A:
			N[dN[j[i]]]+=1
		N=[float(item)/sum(N) for item in N]
		N=[round(item,4) for item in N]
		N=[item+pc for item in N]
		N=[float(item)/sum(N) for item in N]
		N=[round(item,4) for item in N]
		N=[str(item) for item in N]
		N=" ".join(N)
		mat.append(N)
	return mat
	
	
#=======================================================================
# get_matrix based on input Xmer hits
# A: motif Hits

def get_matrix_Extend(A,a,b,x,PP):
	dN={"A":0,"C":1,"G":2,"T":3}
	pc=0.01
	mat=[]
	lm=(X-x)/2
	rm=(X+x)/2
	#======================================
	LM=[item[a:lm] for item in A]
	RM=[item[rm:b] for item in A]
	MM=[item[lm:rm] for item in A]
	PLM=PP[a:lm]
	PRM=PP[rm:b]
	#===========================================
	# handling left
	matL=[]
	for i in PLM:
		ip=i.index(min(i))
		IR=[0,0,0,0]
		IR[ip]=1
		matL.append(IR)
	#============================================
	# handling right
	matR=[]
	for i in PRM:
		ip=i.index(min(i))
		IR=[0,0,0,0]
		IR[ip]=1
		matR.append(IR)
	
	#=============================================
	# handling x-mer
	matX=[]
	
	for i in range(x):
		N=[0,0,0,0]
		for j in MM:
			N[dN[j[i]]]+=1
		matX.append(N)
	#==================================================
	mat=matL+matX+matR
	for i in range(len(mat)):
		mat[i]=[float(item)/sum(mat[i]) for item in mat[i]]
		mat[i]=[round(item,4) for item in mat[i]]
		mat[i]=[item+pc for item in mat[i]]
		mat[i]=[float(item)/sum(mat[i]) for item in mat[i]]
		mat[i]=[round(item,4) for item in mat[i]]
		mat[i]=[str(item) for item in mat[i]]
		mat[i]=" ".join(mat[i])
	#pdb.set_trace()
	return mat

	
#=======================================================================
# check whether 2 motifs are similar

def sub_mot(a,b):
	if len(a)>len(b):
		return False
	elif len(a)<=len(b):
		for i in range(len(b)-len(a)+1):
			flag=0
			#pdb.set_trace()
			for j in range(len(a)):
				bc=dN[b[i+j]]
				ac=dN[a[j]]
				ov=[item for item in ac if item not in bc]
				if len(ov)!=0:
					flag+=1
			if flag<=1:
				return True
		return False	
		
	
def share_mot(a,b):
	if len(a)<len(b):
		t=b
		b=a
		a=t
	a1=a[1:]
	a2=a[:-1]
	if sub_mot(a,b):
		return a
	if sub_mot(a1,b):
		return a1
	elif sub_mot(a2,b):
		return a2
	else:
		return ""
		
#======================================================================
# getting the shorten postion in the x-mer obtained 	
def get_pos(ss,xm,x,X):
	xm=xm[(X-x)/2:(X+x)/2]
	MM=[]
	for i in range(len(xm)-len(ss)+1):
		MM.append(mismatch(ss,xm[i:i+len(ss)]))
	return MM.index(min(MM))+(X-x)/2		

#=======================================================================
# calculating the mismatches among 2 sequences
def mismatch(a,b):
	miscount=0
	for i in range(len(a)):
		if a[i]!=b[i]:
			miscount+=1
	return miscount
	
	
########################################################################


########################################################################
# processing input files
# creating dictionaries

#========================================================================
# read in files
#mot: .motifs
#hit: motif hits
#pos: motif hits positions
#seq: 

mot=readMot(sys.argv[1])
hit=readMotHit(sys.argv[3])
pos=readMotHit(sys.argv[4])
seq=readSeq(sys.argv[5])

#=======================================================================
dm={} # dictionary of moitfs -> targets 
ct=0
for i in mot:
	mi=i[0][1:]
	tar=motTarget(mi,hit)
	dm[mi]=tar
	ct+=1



#=======================================================================
dmm={} # dictionary of motifs-> co-working motifs 


for i in mot:
	dmm[i[0][1:]]=[0]*len(mot)

f=open(sys.argv[2],'r')
ct=0
for line in f:
	line=line.strip().split()[:-2]
	line=[item[1:] for item in line]
	for j in line:
		for k in line:
			dmm[j][int(k)]=1
			
	ct+=1
f.close()
#=======================================================================
# processing the hash table dmm

for i in dmm.keys():
	val=dmm[i]
	tmp=[]
	for j in range(len(val)):
		if val[j]==1:
			tmp.append(mot[j][0][1:])
	dmm[i]=tmp

#=======================================================================
# getting targets of motifs (in all motif modules having the motifs)

dmc={}
ct=0
for i in mot:
	mi=i[0][1:]
	tarmc=motTargetMC(mi,dm,dmm)
	dmc[mi]=tarmc
	ct+=1

#=======================================================================
########################################################################

########################################################################
# get X-mers 
########################################################################
#======================================================================
## Hits for each motif in target genes
P=[]
XMID=[]
for i in mot:
	mi=i[0][1:]
	if dmc.has_key(mi)==True:
		XMID.append(mi)
		T=dmc[mi]
		posi=[]
		for j in T:
			indi_j=[k for k in range(len(hit[j])) if hit[j][k]==mi]
			pos_ij=[pos[j][k] for k in indi_j]
			pos_ij=[str(j)+","+item for item in pos_ij]
			#pdb.set_trace()
			posi+=pos_ij
		P.append(posi)
#======================================================================
## 24-mers
X=24 # 24-mers
x=int(mot[0][1]) # length of motif
XM=[]
for i in range(len(P)):
	XM_i=[]
	for j in range(len(P[i])):
		sid=int(P[i][j].split(",")[0])
		loc=int(P[i][j].split(",")[1])
		xm=seq[sid][int(loc-(X-x)/2):int(loc+x+(X-x)/2)]
		XM_i.append(xm)
	XM.append(XM_i)

#=======================================================================
#filtering XM-xmers, so that some hits (length <X, or containing "N" 
#will be removed

for i in range(len(XM)):
	XM[i]=[item for item in XM[i] if len(item)==X]
	for j in range(len(XM[i])):
		if "N" in XM[i][j]:
			XM[i][j]=""
	XM[i]=[item for item in XM[i] if item!=""]

#######################################################################
# Extension starts here!

#======================================================================
P=[]
cut=1e-12 # pvalue cut
cutp=0.5
bg=[0.247,0.251,0.254,0.248]
d={"A":0,"C":1,"G":2,"T":3}

extend_mot=[]
extend_ID=[]

unextend_mot=[]
unextend_ID=[]
ct=0
maxLength=14 # max length of motif
for i in XM:
	Pl=[] # for left
	Ptl=[]
	Pr=[] # for right
	Ptr=[]
	Ph=[] # for x-mer
	G=len(i)
	if G>0:
		## handling left x-mer
		for j in range(int((X-x)/2)):
			A=[0,0,0,0]
			for k in i:
				A[d[k[j]]]+=1
			pv_a=max(0,1-pbinom(A[0]-1,G,bg[0]))
			pv_c=max(0,1-pbinom(A[1]-1,G,bg[1]))
			pv_g=max(0,1-pbinom(A[2]-1,G,bg[2]))
			pv_t=max(0,1-pbinom(A[3]-1,G,bg[3]))
			pv=[pv_a,pv_c,pv_g,pv_t]
			Ptl.append([float(item)/sum(A) for item in A])
			Pl.append(pv)
			
		## handling right x-mer
		
		for j in range(int((X+x)/2),int(X)):
			A=[0,0,0,0]
			for k in i:
				A[d[k[j]]]+=1
			pv_a=max(0,1-pbinom(A[0]-1,G,bg[0]))
			pv_c=max(0,1-pbinom(A[1]-1,G,bg[1]))
			pv_g=max(0,1-pbinom(A[2]-1,G,bg[2]))
			pv_t=max(0,1-pbinom(A[3]-1,G,bg[3]))
			pv=[pv_a,pv_c,pv_g,pv_t]	
			Pr.append(pv)
			Ptr.append([float(item)/sum(A) for item in A])
			
			
		## handling x-mer
		
		for j in range(int((X-x)/2),int((X+x)/2)):
			A=[0,0,0,0]
			for k in i:
				A[d[k[j]]]+=1
			pv_a=max(0,1-pbinom(A[0]-1,G,bg[0]))
			pv_c=max(0,1-pbinom(A[1]-1,G,bg[1]))
			pv_g=max(0,1-pbinom(A[2]-1,G,bg[2]))
			pv_t=max(0,1-pbinom(A[3]-1,G,bg[3]))
			pv=[pv_a,pv_c,pv_g,pv_t]	
			Ph.append(pv)
		
		## motif extension#################################################
		dN={"A":"A","C":"C","G":"G","T":"T","AG":"R","CT":"Y","GT":"K","AC":"M","CG":"S","AT":"W","CGT":"B","AGT":"D","ACT":"H","ACG":"V","ACGT":"N"}
		## left extension
		prefix_con=[]
		for j in range(len(Pl)-1,-1,-1):
			letter=""
			flag=0
			if Pl[j][0]<cut:
				letter+="A"
				flag+=1
			if Pl[j][1]<cut:
				letter+="C"
				flag+=1
			if Pl[j][2]<cut:
				letter+="G"
				flag+=1
			if Pl[j][3]<cut:
				letter+="T"
				flag+=1
			if max(Ptl[j])<cutp:
				flag=0
			if flag==0:
				break
			prefix_con.append(dN[letter])
		prefix_con=prefix_con[::-1]
		#pdb.set_trace()
		## right extension
		suffix_con=[]
		for j in range(len(Pr)):
			letter=""
			flag=0
			if Pr[j][0]<cut:
				letter+="A"
				flag+=1
			if Pr[j][1]<cut:
				letter+="C"
				flag+=1
			if Pr[j][2]<cut:
				letter+="G"
				flag+=1
			if Pr[j][3]<cut:
				letter+="T"
				flag+=1
			if max(Ptr[j])<cutp:
				flag=0
			if flag==0:
				break
			suffix_con.append(dN[letter])
		# x-mer consensus
		hit_con=[]
		for j in range(len(Ph)):
			letter=""
			flag=0
			if Ph[j][0]<cut:
				letter+="A"
				flag+=1
			if Ph[j][1]<cut:
				letter+="C"
				flag+=1
			if Ph[j][2]<cut:
				letter+="G"
				flag+=1
			if Ph[j][3]<cut:
				letter+="T"
				flag+=1
			if flag==0:
				letter+="ACGT"
			hit_con.append(dN[letter])
		
		sp=(X-x)/2-len(prefix_con) # extension starts
		ep=(X+x)/2+len(suffix_con) # extension ends 
		
		if ep-sp>maxLength:
			ep=sp+maxLength
		XMIS=[item[sp:ep] for item in i]
		if len(XMIS)>0:
			if ep-sp>x:
				mot_i=get_matrix(XMIS)
				extend_mot.append(mot_i)
				extend_ID.append(XMID[ct])
			else:
				con=prefix_con+hit_con+suffix_con
				unextend_mot.append(con)
				unextend_ID.append(XMID[ct])
	ct+=1
########################################################################

########################################################################
# shorten starts here!
dN={"A":"A","C":"C","G":"G","T":"T","R":"AG","Y":"CT","K":"GT","M":"AC","S":"CG","W":"AT","B":"CGT","D":"AGT","H":"ACT","V":"ACG","N":"ACGT"}


minwidth=6 # mininal length of motif

#pdb.set_trace()
for i in range(len(unextend_mot)-1):
		for j in range(i+1,len(unextend_mot)):
			if (unextend_mot[i]!=[]) and (unextend_mot[j]!=[]):
				cs=share_mot(unextend_mot[i],unextend_mot[j])
				if len(cs)>=minwidth:
					mi=unextend_ID[i].split(",")[0]
					mj=unextend_ID[j].split(",")[0]
					mc_share=[item for item in dmm[mi]  if item in dmm[mj]]
					mc_share=[item for item in mc_share if item!=mi and item!=mj]
					if len(mc_share)>0:
						unextend_mot[i]=cs
						unextend_mot[j]=[]
						unextend_ID[i]+=","+unextend_ID[j]
						unextend_ID[j]=[]
unextend_mot=[item for item in unextend_mot if item!=[]]
unextend_ID=[item for item in unextend_ID if item!=[]]

#======================================================================
# get matrix

unextend_mot_mat=[]
for i in range(len(unextend_ID)):
	mi=unextend_ID[i].split(",")[0]
	ind=XMID.index(mi)
	xmi=XM[ind]
	ss=unextend_mot[i]
	ind_i=get_pos(ss,xmi[0],x,X)
	xmi=[item[ind_i:ind_i+len(ss)] for item in xmi]
	mi=get_matrix(xmi)
	unextend_mot_mat.append(mi)
del unextend_mot
#=======================================================================
#  outputing... 

ID=extend_ID+unextend_ID
dID={} # dictionary for old ID-> new ID
Score=[] # MDScan score 
for i in range(len(ID)):
	IDI=ID[i].split(",")
	for j in IDI:
		dID[j]=i
	Inds=XMID.index(IDI[0])
	Score.append(mot[Inds][-1])
mot=extend_mot+unextend_mot_mat
print ("outputing motifs")

for i in range(len(mot)):
	mot[i]=">M"+str(i)+"\t"+Score[i]+"\n"+"\n".join(mot[i])
mot="\n".join(mot)

outputdir=sys.argv[6]
if os.path.exists(outputdir)==False:
	os.mkdir(outputdir)
FN=sys.argv[1].split("/")[-1]
f=open(outputdir+"/"+FN+".adjusted",'w')
f.write(mot)
f.close()

print("outputing motif modules...")

FN=sys.argv[2].split("/")[-1]
f=open(sys.argv[2],'r')
g=open(outputdir+"/"+FN+".adjusted",'a')
for line in f:
	line=line.strip().split()
	score=line[-2:]
	line=line[:-2]
	line=[item[1:] for item in line]
	line=["M"+str(dID[item]) for item in line]
	line=" ".join(line)+"\t"+" ".join(score)
	g.write(line+"\n")
g.close()
f.close()



